"""
DaWiQ - Dataclass Widget for Qt
===============================

DaWiQ is a package to build Qt widgets from dataclasses.

"""

from .version import __version__  # noqa


from .fieldwidgets import (
    BoolCheckBox,
    EmptyIntValidator,
    IntLineEdit,
    EmptyFloatValidator,
    FloatLineEdit,
    StrLineEdit,
    EnumComboBox,
    TupleGroupBox,
)
from .datawidget import (
    DataWidget,
    type2Widget,
    dataclass2Widget,
)
from .multitype import (
    DataclassStackedWidget,
    DataclassTabWidget,
)
from .delegate import (
    convertFromQt,
    convertToQt,
    highlightEmptyField,
    DataclassDelegate,
    DataclassMapper,
)


__all__ = [
    "BoolCheckBox",
    "EmptyIntValidator",
    "IntLineEdit",
    "EmptyFloatValidator",
    "FloatLineEdit",
    "StrLineEdit",
    "EnumComboBox",
    "TupleGroupBox",
    "DataWidget",
    "type2Widget",
    "dataclass2Widget",
    "DataclassStackedWidget",
    "DataclassTabWidget",
    "convertFromQt",
    "convertToQt",
    "highlightEmptyField",
    "DataclassDelegate",
    "DataclassMapper",
]



"""
Type annotations
================

:mod:`dawiq.typing` provides special type annotations.
"""

from typing import Protocol, Dict, Any
from .qt_compat import QtCore


__all__ = [
    "DataclassProtocol",
    "FieldWidgetProtocol",
]


class DataclassProtocol(Protocol):
    """Type annotation for dataclass type object."""

    # https://stackoverflow.com/a/55240861/11501976
    __dataclass_fields__: Dict


class FieldWidgetProtocol(Protocol):
    """Type annotation for field widget object."""

    fieldValueChanged: QtCore.Signal
    fieldEdited: QtCore.Signal

    def fieldValue(self) -> Any:
        """
        Value that the field widget represents.

        This is the API for the delegate to get the data from the field widget.
        For example, in :class:`BoolCheckBox <dawiq.fieldwidgets.BoolCheckBox>`
        this method converts the current check state to :class:`bool` and
        returns.

        If the field value is :obj:`None`, it typically indicates that the widget
        is empty and the delegate should specially handle it. One of the few
        exceptions is :class:`BoolCheckBox <dawiq.fieldwidgets.BoolCheckBox>`
        where :obj:`None` indicates partially checked state for fuzzy boolean
        value.

        When the field value is changed, :attr:`fieldValueChanged` signal must
        emit the new value. When the field is edited by user, :attr:`fieldEdited`
        signal must be emitted with empty argument.

        """
        ...

    def setFieldValue(self, value: Any):
        """
        Set the value to the field widget.

        This method is the API for the delegate to set the data to the widget.
        For example, in :class:`BoolCheckBox <dawiq.fieldwidgets.BoolCheckBox>`
        this method converts :class:`bool` to :obj:`Qt.CheckState` and sets to
        the widget.

        For valid *value*, its type must be strictly checked and :obj:`TypeError`
        must be raised on invalid input.

        """
        ...

    def fieldName(self) -> str:
        """
        Name of the field.

        It is recommended to make this name visible in the widget without
        affecting the data value. Placeholder text of the line edit or title of
        the group box are good examples.

        """
        ...

    def setFieldName(self, name: str):
        """Set the name of the field."""
        ...

    def setRequired(self, required: bool):
        """
        Set if *self* represents a required field.

        If *required* is True, it indicates that the field is mandatory. On such
        case, this methods checks the :meth:`fieldValue` of *self* and sets
        ``requiresFieldValue`` property of the editor widget. If the data value of
        required field is missing, the property is set to be True.

        .. code-block:: python

            widget.setProperty("requiresFieldValue", True)
            widget.style().unpolish(widget)
            widget.style().polish(widget)

        The editor widget is usually *self*, but there are exceptions such as
        :class:`TupleGroupBox`. Note that the widget needs to be re-polished.

        Style sheet can be set to highlight the required field with empty widget.
        Common way is to set the style sheet of :class:`QApplication` before
        starting the app.

        .. code-block:: python

            qApp.setStyleSheet(
                "*[requiresFieldValue=true]{border: 1px solid red}"
            )

        Notes
        =====

        This method is designed to be called by the :class:`DataclassDelegate`,
        not directly by the user.

        """
        ...



"""
Data field widgets
==================

:mod:`dawiq.fieldwidgets` provides widgets to represent the fields of the
dataclass. Widgets are compatible to :class:`dawiq.typing.FieldWidgetProtocol`.
"""

from .qt_compat import QtCore, QtWidgets, QtGui
from enum import Enum
from typing import Optional, Union, Tuple, TypeVar, Type, Any
from .typing import FieldWidgetProtocol


__all__ = [
    "BoolCheckBox",
    "EmptyIntValidator",
    "IntLineEdit",
    "EmptyFloatValidator",
    "FloatLineEdit",
    "StrLineEdit",
    "EnumComboBox",
    "TupleGroupBox",
]


class BoolCheckBox(QtWidgets.QCheckBox):
    """
    Checkbox for fuzzy boolean value.

    Check state of the box represents the field value. If the box is checked, the
    value is True. If unchecked, the value is False.

    If tristate is enabled, setting ``None`` as the field value sets the state as
    ``Qt.PartiallyChecked``. If tristate is disabled, ``None`` is treated as
    ``False`` and unchecks the box.

    """

    fieldValueChanged = QtCore.Signal(object)
    fieldEdited = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.stateChanged.connect(self._onStateChange)
        self.clicked.connect(self.fieldEdited)

    def fieldValue(self) -> Optional[bool]:
        checkstate = self.checkState()
        if checkstate == QtCore.Qt.CheckState.Checked:
            state = True
        elif checkstate == QtCore.Qt.CheckState.Unchecked:
            state = False
        else:
            state = None
        return state

    def setFieldValue(self, value: Optional[bool]):
        if value is None and not self.isTristate():
            value = False
        if value is True:
            state = QtCore.Qt.CheckState.Checked
        elif value is False:
            state = QtCore.Qt.CheckState.Unchecked
        elif value is None:
            state = QtCore.Qt.CheckState.PartiallyChecked
        else:
            raise TypeError(
                f"BoolCheckBox data must be True, False or None, not {type(value)}"
            )
        self.setCheckState(state)

    def _onStateChange(self, checkstate: Union[int, QtCore.Qt.CheckState]):
        checkstate = QtCore.Qt.CheckState(checkstate)
        if checkstate == QtCore.Qt.CheckState.Checked:
            state = True
        elif checkstate == QtCore.Qt.CheckState.Unchecked:
            state = False
        else:
            state = None
        self.fieldValueChanged.emit(state)

    def fieldName(self) -> str:
        return self.text()

    def setFieldName(self, name: str):
        self.setText(name)
        self.setToolTip(name)

    def setRequired(self, required: bool):
        # Check box is always occupied
        pass


class EmptyIntValidator(QtGui.QIntValidator):
    """Validator which accpets integer and empty string"""

    def validate(self, input: str, pos: int) -> Tuple[QtGui.QValidator.State, str, int]:
        state, ret_input, ret_pos = super().validate(input, pos)
        if not input:
            state = QtGui.QValidator.State.Acceptable
        return (state, ret_input, ret_pos)


class IntLineEdit(QtWidgets.QLineEdit):
    """
    Line edit for integer value.

    If the text is not empty, the field value is the integer that the text is
    converted to. If the line edit is empty or the text cannot be converted to
    integer, ``None`` is the field value. Setting ``None`` as field value clears
    the line edit.

    """

    fieldValueChanged = QtCore.Signal(object)
    fieldEdited = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setValidator(EmptyIntValidator(self))

        self.textChanged.connect(self._onTextChange)
        self.editingFinished.connect(self.fieldEdited)

    def fieldValue(self) -> Optional[int]:
        text = self.text()
        if not text:
            val: Optional[int] = None
        else:
            try:
                val = int(text)
            except ValueError:
                val = None
        return val

    def setFieldValue(self, value: Optional[int]):
        if value is None:
            txt = ""
        elif isinstance(value, int):
            txt = str(int(value))
        else:
            raise TypeError(f"IntLineEdit value must be int, not {type(value)}")
        self.setText(txt)

    def _onTextChange(self, text: str):
        if not text:
            val: Optional[int] = None
        else:
            try:
                val = int(text)
            except ValueError:
                val = None
        self.fieldValueChanged.emit(val)

    def fieldName(self) -> str:
        return self.placeholderText()

    def setFieldName(self, name: str):
        self.setPlaceholderText(name)
        self.setToolTip(name)

    def setRequired(self, required: bool):
        if required and self.fieldValue() is None:
            requires = True
        else:
            requires = False
        if self.property("requiresFieldValue") != requires:
            self.setProperty("requiresFieldValue", requires)
            self.style().unpolish(self)
            self.style().polish(self)


class EmptyFloatValidator(QtGui.QDoubleValidator):
    """Validator which accpets float and empty string"""

    def validate(self, input: str, pos: int) -> Tuple[QtGui.QValidator.State, str, int]:
        state, ret_input, ret_pos = super().validate(input, pos)
        if not input:
            state = QtGui.QValidator.State.Acceptable
        return (state, ret_input, ret_pos)


class FloatLineEdit(QtWidgets.QLineEdit):
    """
    Line edit for float value.

    If the text is not empty, the field value is the float that the text is
    converted to. If the line edit is empty or the text cannot be converted to
    float, ``None`` is the field value. Setting ``None`` as field value clears
    the line edit.
    """

    fieldValueChanged = QtCore.Signal(object)
    fieldEdited = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setValidator(EmptyFloatValidator(self))

        self.textChanged.connect(self._onTextChange)
        self.editingFinished.connect(self.fieldEdited)

    def fieldValue(self) -> Optional[float]:
        text = self.text()
        if not text:
            val: Optional[float] = None
        else:
            try:
                val = float(text)
            except ValueError:
                val = None
        return val

    def setFieldValue(self, value: Optional[float]):
        if value is None:
            txt = ""
        elif isinstance(value, float):
            txt = str(float(value))
        else:
            raise TypeError(f"FloatLineEdit value must be float, not {type(value)}")
        self.setText(txt)

    def _onTextChange(self, text: str):
        if not text:
            val: Optional[float] = None
        else:
            try:
                val = float(text)
            except ValueError:
                val = None
        self.fieldValueChanged.emit(val)

    def fieldName(self) -> str:
        return self.placeholderText()

    def setFieldName(self, name: str):
        self.setPlaceholderText(name)
        self.setToolTip(name)

    def setRequired(self, required: bool):
        if required and self.fieldValue() is None:
            requires = True
        else:
            requires = False
        if self.property("requiresFieldValue") != requires:
            self.setProperty("requiresFieldValue", requires)
            self.style().unpolish(self)
            self.style().polish(self)


class StrLineEdit(QtWidgets.QLineEdit):
    """
    Line edit for string value.

    If the line edit is empty, field data is empty string. Setting ``None`` as
    the field value clears the widget.

    """

    fieldValueChanged = QtCore.Signal(str)
    fieldEdited = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.textChanged.connect(self._onTextChange)
        self.editingFinished.connect(self.fieldEdited)

    def fieldValue(self) -> str:
        return self.text()

    def setFieldValue(self, value: Optional[str]):
        if value is None:
            txt = ""
        elif isinstance(value, str):
            txt = str(value)  # type: ignore[assignment]
        else:
            raise TypeError(f"StrLineEdit data must be str, not {type(value)}")
        self.setText(txt)

    def _onTextChange(self, text: str):
        self.fieldValueChanged.emit(text)

    def fieldName(self) -> str:
        return self.placeholderText()

    def setFieldName(self, name: str):
        self.setPlaceholderText(name)
        self.setToolTip(name)

    def setRequired(self, required: bool):
        if required and self.fieldValue() is None:
            requires = True
        else:
            requires = False
        if self.property("requiresFieldValue") != requires:
            self.setProperty("requiresFieldValue", requires)
            self.style().unpolish(self)
            self.style().polish(self)


T = TypeVar("T", bound="EnumComboBox")


class EnumComboBox(QtWidgets.QComboBox):
    """
    Combo box for :class:`enum.Enum` type.

    Standard way to construct this widget is by :meth:`fromEnum` class method.
    N-th item contains N-th member of the Enum as its data.

    Enum instance is stored in item data. Field value is the data of currently
    activated item. If the current index is -1, field value is ``None``.

    """

    fieldValueChanged = QtCore.Signal(object)
    fieldEdited = QtCore.Signal()

    @classmethod
    def fromEnum(cls: Type[T], enum: Type[Enum]) -> T:
        obj = cls()
        for e in enum:
            obj.addItem(e.name, userData=e)
        obj.setCurrentIndex(-1)
        return obj

    def __init__(self, parent=None):
        super().__init__(parent)
        self.currentIndexChanged.connect(self._onCurrentIndexChange)
        self.activated.connect(self.fieldEdited)

    def fieldValue(self) -> Optional[Enum]:
        index = self.currentIndex()
        if index == -1:
            ret = None
        else:
            ret = self.itemData(index)
        return ret

    def setFieldValue(self, value: Optional[Enum]):
        if value is None:
            index = -1
        elif isinstance(value, Enum):
            index = self.findData(value)
        else:
            raise TypeError(f"EnumComboBox data must be Enum, not {type(value)}")
        self.setCurrentIndex(index)

    def _onCurrentIndexChange(self, index: int):
        data = self.itemData(index)
        self.fieldValueChanged.emit(data)

    def fieldName(self) -> str:
        return self.placeholderText()

    def setFieldName(self, name: str):
        self.setPlaceholderText(name)
        self.setToolTip(name)

    def setRequired(self, required: bool):
        if required and self.fieldValue() is None:
            requires = True
        else:
            requires = False
        if self.property("requiresFieldValue") != requires:
            self.setProperty("requiresFieldValue", requires)
            self.style().unpolish(self)
            self.style().polish(self)


V = TypeVar("V", bound="TupleGroupBox")


class TupleGroupBox(QtWidgets.QGroupBox):
    """
    Group box for tuple with fixed length.

    This is the group box which contains field widgets as subwidgets. Field value
    is the tuple of subwidgets values.

    """

    fieldValueChanged = QtCore.Signal(tuple)
    fieldEdited = QtCore.Signal()

    def __init__(
        self,
        orientation: QtCore.Qt.Orientation = QtCore.Qt.Orientation.Horizontal,
        parent=None,
    ):
        super().__init__(parent)
        self._orientation = orientation

        if orientation == QtCore.Qt.Orientation.Vertical:
            layout = QtWidgets.QVBoxLayout()
        elif orientation == QtCore.Qt.Orientation.Horizontal:
            layout = QtWidgets.QHBoxLayout()
        else:
            raise TypeError(f"Invalid orientation: {orientation}")
        self.setLayout(layout)

    def orientation(self) -> QtCore.Qt.Orientation:
        """Orientation to stack the subwidgets."""
        return self._orientation

    def count(self) -> int:
        """Number of subwidgets."""
        return self.layout().count()

    def widget(self, index: int) -> Optional[FieldWidgetProtocol]:
        """
        Returns the subwidget at the given index, or None for invalid index.
        """
        item = self.layout().itemAt(index)
        if item is not None:
            item = item.widget()
        return item

    def insertWidget(
        self,
        index: int,
        widget: FieldWidgetProtocol,
        stretch: int = 0,
        alignment: QtCore.Qt.AlignmentFlag = QtCore.Qt.AlignmentFlag(0),
    ):
        """Insert the widget to layout and connect the signals."""
        for i in range(self.count()):
            w = self.widget(i)
            if w is None:
                break
        widget.fieldValueChanged.connect(self._onSubfieldValueChange)
        widget.fieldEdited.connect(self.fieldEdited)
        self.layout().insertWidget(index, widget, stretch, alignment)

    def addWidget(
        self,
        widget: FieldWidgetProtocol,
        stretch: int = 0,
        alignment: QtCore.Qt.AlignmentFlag = QtCore.Qt.AlignmentFlag(0),
    ):
        """Add the widget to layout and connect the signals."""
        for i in range(self.count()):
            w = self.widget(i)
            if w is None:
                break
        widget.fieldValueChanged.connect(self._onSubfieldValueChange)
        widget.fieldEdited.connect(self.fieldEdited)
        self.layout().addWidget(widget, stretch, alignment)

    def removeWidget(self, widget: FieldWidgetProtocol):
        """Remove the widget from layout and disconnect the signals."""
        for i in range(self.count()):
            w = self.widget(i)
            if w is None:
                break
            if w == widget:
                widget.fieldValueChanged.disconnect(self._onSubfieldValueChange)
                widget.fieldEdited.disconnect(self.fieldEdited)
                break
        self.layout().removeWidget(widget)

    def fieldValue(self) -> tuple:
        ret = []
        for i in range(self.count()):
            widget = self.widget(i)
            if widget is None:
                break
            ret.append(widget.fieldValue())
        return tuple(ret)

    def setFieldValue(self, value: Optional[tuple]):
        if value is None:
            value = tuple(None for _ in range(self.count()))
        elif isinstance(value, tuple):
            pass
        else:
            raise TypeError(f"TupleGroupBox value must be tuple, not {type(value)}")

        for i in range(self.count()):
            widget = self.widget(i)
            if widget is None:
                break
            widget.fieldValueChanged.disconnect(self._onSubfieldValueChange)
            widget.setFieldValue(value[i])
            widget.fieldValueChanged.connect(self._onSubfieldValueChange)
        self.fieldValueChanged.emit(value)

    def _onSubfieldValueChange(self, value: Any):
        self.fieldValueChanged.emit(self.fieldValue())

    def fieldName(self) -> str:
        return self.title()

    def setFieldName(self, name: str):
        self.setTitle(name)
        self.setToolTip(name)

    def setRequired(self, required: bool):
        """Recursively set *required* to all subwidgets."""
        for i in range(self.count()):
            widget = self.widget(i)
            if widget is None:
                continue
            widget.setRequired(required)



"""
Dataclass delegate
==================

"""

import dataclasses
from .qt_compat import QtWidgets, TypeRole, DataRole
from .datawidget import DataWidget
from .multitype import DataclassStackedWidget, DataclassTabWidget
from .typing import DataclassProtocol
from typing import Type, Dict, Any, Optional


__all__ = [
    "convertFromQt",
    "convertToQt",
    "highlightEmptyField",
    "DataclassDelegate",
    "DataclassMapper",
]


def convertFromQt(
    dcls: Type[DataclassProtocol],
    data: Dict[str, Any],
    ignoreMissing=True,
) -> Dict[str, Any]:
    """
    Convert the dict from :class:`DataWidget` to structured dict for dataclass.

    If the field value does not exist in *data* or is :obj:`None`, it is
    considered to be absent. If *ignoreMissing* is True, the absent value is not
    included in the resulting dictionary. Else, the default value of the field is
    used if there is any.

    Field may define ``fromQt_converter`` metadata to convert the widget data to
    field data. It is a unary callable which takes the widget data and returns
    the field data.

    Examples
    ========

    ``None`` is considered as missing value.

    >>> from dataclasses import dataclass, field
    >>> from dawiq.delegate import convertFromQt
    >>> from typing import Optional
    >>> @dataclass
    ... class Cls1:
    ...     a: int
    ...     b: int = 10
    ...     c: Optional[int] = None
    ...     d: list = field(default_factory=list)
    >>> convertFromQt(Cls1, {})
    {}
    >>> convertFromQt(Cls1, dict(a=1, b=None, c=None))
    {'a': 1}

    *ignoreMissing* controls whether the default value should be used.

    >>> convertFromQt(Cls1, {}, ignoreMissing=False)
    {'b': 10, 'c': None, 'd': []}

    Nested dataclasses are recusively converted.

    >>> @dataclass
    ... class Cls2:
    ...     x: int = 20
    ...     y: Cls1 = field(default_factory=lambda: Cls1(a=5))
    >>> convertFromQt(Cls2, {}, ignoreMissing=False)
    {'x': 20, 'y': {'a': 5, 'b': 10, 'c': None, 'd': []}}

    ``fromQt_converter`` metadata converts the data from the widet.

    >>> def conv(arg):
    ...     return (arg,)
    >>> @dataclass
    ... class Cls3:
    ...     x: tuple = field(default=(1,), metadata=dict(fromQt_converter=conv))
    >>> convertFromQt(Cls3, dict(x=1))
    {'x': (1,)}

    """
    # Return value is not dataclass but dictionary because necessary fields might
    # be missing from the widget.
    ret: Dict[str, Any] = {}
    for f in dataclasses.fields(dcls):
        val = data.get(f.name, None)
        t = f.type

        if val is None:
            if ignoreMissing:
                continue
            val = (
                f.default_factory()
                if f.default_factory is not dataclasses.MISSING
                else f.default
            )
            if val is dataclasses.MISSING:
                if dataclasses.is_dataclass(t):
                    val = convertFromQt(t, {}, ignoreMissing)
                else:
                    continue
            if dataclasses.is_dataclass(val) and not isinstance(val, type):
                val = dataclasses.asdict(val)

        else:
            if dataclasses.is_dataclass(t):
                val = convertFromQt(t, val, ignoreMissing)
            converter = f.metadata.get("fromQt_converter", None)
            if converter is not None:
                val = converter(val)
        ret[f.name] = val
    return ret


def convertToQt(
    dcls: Type[DataclassProtocol],
    data: Dict[str, Any],
    ignoreMissing=True,
) -> Dict[str, Any]:
    """
    Convert structured dict from dataclass to dict for :class:`DataWidget`.

    If the field value does not exist in *data* or is :obj:`None`, it is
    considered to be absent. If *ignoreMissing* is True, ``None`` is used as the
    placeholder in the resulting dictionary. Else, the default value of the field
    is used if there is any.

    Field may define ``toQt_converter`` metadata to convert the field data to
    widget data. It is a unary callable which takes the field data and returns
    the widget data.

    Examples
    ========

    ``None`` is considered as missing value.

    >>> from dataclasses import dataclass, field
    >>> from dawiq.delegate import convertToQt
    >>> from typing import Optional
    >>> @dataclass
    ... class Cls1:
    ...     a: int
    ...     b: int = 10
    ...     c: Optional[int] = None
    ...     d: list = field(default_factory=list)
    >>> convertToQt(Cls1, dict(a=1, b=2, c=3))
    {'a': 1, 'b': 2, 'c': 3, 'd': None}
    >>> convertToQt(Cls1, {})
    {'a': None, 'b': None, 'c': None, 'd': None}

    *ignoreMissing* controls whether the default value should be used.

    >>> convertToQt(Cls1, {}, ignoreMissing=False)
    {'a': None, 'b': 10, 'c': None, 'd': []}

    Nested dataclasses are recusively converted.

    >>> @dataclass
    ... class Cls2:
    ...     x: int = 20
    ...     y: Cls1 = field(default_factory=lambda: Cls1(a=5))
    >>> convertToQt(Cls2, {}, ignoreMissing=False)
    {'x': 20, 'y': {'a': 5, 'b': 10, 'c': None, 'd': []}}

    ``toQt_converter`` metadata converts the data to the widget.

    >>> @dataclass
    ... class Cls3:
    ...     x: int = field(metadata=dict(toQt_converter=lambda tup: tup[0]))
    >>> convertToQt(Cls3, dict(x=(1,)))
    {'x': 1}

    """
    ret: Dict[str, Any] = {}
    for f in dataclasses.fields(dcls):
        val = data.get(f.name, None)
        t = f.type

        if val is None:
            if ignoreMissing:
                ret[f.name] = None
                continue
            val = (
                f.default_factory()
                if f.default_factory is not dataclasses.MISSING
                else f.default
            )
            if val is dataclasses.MISSING:
                if dataclasses.is_dataclass(t):
                    val = convertToQt(t, {}, ignoreMissing)
                else:
                    ret[f.name] = None
                    continue
            if dataclasses.is_dataclass(val) and not isinstance(val, type):
                val = dataclasses.asdict(val)

        else:
            if dataclasses.is_dataclass(t):
                val = convertToQt(t, val, ignoreMissing)
            converter = f.metadata.get("toQt_converter", None)
            if converter is not None:
                val = converter(val)
        ret[f.name] = val
    return ret


def highlightEmptyField(editor: DataWidget, dcls: Optional[Type[DataclassProtocol]]):
    """Recursively highlight the empty field whose data is required."""
    if dcls is None:
        editor.setRequired(False)
    else:
        # get field widgets from *editor*
        field_widgets = {}
        for i in range(editor.count()):
            widget = editor.widget(i)
            if widget is None:
                continue
            field_widgets[widget.fieldName()] = widget

        # if the field does not have default value, the field is required.
        for f in dataclasses.fields(dcls):
            widget = field_widgets.pop(f.name, None)
            if widget is None:  # no widget for field
                continue
            default = (
                f.default_factory()
                if f.default_factory is not dataclasses.MISSING
                else f.default
            )
            required = default is dataclasses.MISSING
            if isinstance(widget, DataWidget):
                if required and dataclasses.is_dataclass(f.type):
                    highlightEmptyField(widget, f.type)
                else:
                    widget.setRequired(required)
            else:
                widget.setRequired(required)


class DataclassDelegate(QtWidgets.QStyledItemDelegate):
    """
    Delegate to update the model and the data widget.

    This delegate stores dataclass type and dataclass data to the model, and
    updates the data widget with model data.

    Supported data widgets are:

    * :class:`DataWidget`
    * :class:`DataclassStackedWidget`
    * :class:`DataclassTabWidget`

    Dataclass type is stored to the model with :attr:`TypeRole` as item data role
    and dataclass data is stored with :attr:`DataRole`.

    By default, missing values are not replaced by default values of the fields.
    This is to preserve the intentional empty input by the user. Setting
    :meth:`ignoreMissing` changes this behavior.
    """

    TypeRole = TypeRole
    DataRole = DataRole

    def __init__(self, parent=None):
        super().__init__(parent)
        self._ignoreMissing = True

    def ignoreMissing(self) -> bool:
        """If True, default values are used for missing fields."""
        return self._ignoreMissing

    def setIgnoreMissing(self, val: bool):
        self._ignoreMissing = val

    def setModelData(self, editor, model, index):
        if isinstance(editor, (DataclassStackedWidget, DataclassTabWidget)):
            dcls = editor.currentDataclass()
            self.cacheModelData(model, index, dcls, self.TypeRole)
            self.setModelData(editor.currentWidget(), model, index)

        elif isinstance(editor, DataWidget):
            dcls = model.data(index, role=self.TypeRole)
            data = editor.dataValue()
            if dcls is not None:
                data = convertFromQt(dcls, data, self.ignoreMissing())
            self.cacheModelData(model, index, data, self.DataRole)

        super().setModelData(editor, model, index)

    @classmethod
    def cacheModelData(cls, model, index, value, role):
        """
        Cache the *value* to *model*'s *index* with *role* so that ``submit()``
        method of *model* can permanently store the data.

        Caching API is not defined for general ``QAbstractItemModel``. So by
        default, this method just stores the data to the model without caching.

        Subclass may reimplement this method to actually cache the data to the
        model.

        """
        model.setData(index, value, role)

    def setEditorData(self, editor, index):
        if isinstance(editor, (DataclassStackedWidget, DataclassTabWidget)):
            dcls = index.data(role=self.TypeRole)
            if dcls is not None:
                widgetIndex = editor.indexOfDataclass(dcls)
            else:
                widgetIndex = -1
            editor.setCurrentIndex(widgetIndex)
            self.setEditorData(editor.currentWidget(), index)

        elif isinstance(editor, DataWidget):
            dcls = index.data(role=self.TypeRole)
            data = index.data(role=self.DataRole)
            if data is None:
                data = {}
            if dcls is not None:
                data = convertToQt(dcls, data, self.ignoreMissing())
            editor.setDataValue(data)
            highlightEmptyField(editor, dcls)

        super().setEditorData(editor, index)


class DataclassMapper(QtWidgets.QDataWidgetMapper):
    """
    Mapper between the data widget and the model.

    Supported data widgets are:

    * :class:`DataWidget`
    * :class:`DataclassStackedWidget`
    * :class:`DataclassTabWidget`

    Notes
    =====

    When mapping :class:`DataWidget`, *propertyName* argument of
    :meth:`addMapping` must not be passed.
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setSubmitPolicy(self.SubmitPolicy.ManualSubmit)

    def addMapping(self, widget, section, propertyName=b""):
        if isinstance(widget, DataclassStackedWidget):
            widget.currentDataEdited.connect(self.submit)
        elif isinstance(widget, DataclassTabWidget):
            widget.activated.connect(self.submit)
            widget.currentDataEdited.connect(self.submit)
        elif isinstance(widget, DataWidget):
            widget.dataEdited.connect(self.submit)
        super().addMapping(widget, section, propertyName)

    def removeMapping(self, widget):
        if isinstance(widget, DataclassStackedWidget):
            widget.currentDataEdited.disconnect(self.submit)
        elif isinstance(widget, DataclassTabWidget):
            widget.activated.disconnect(self.submit)
            widget.currentDataEdited.disconnect(self.submit)
        elif isinstance(widget, DataWidget):
            widget.dataEdited.disconnect(self.submit)
        super().removeMapping(widget)



"""
Data widget container
=====================

:mod:`dawiq.multitype` provides widgets that contain multiple :class:`DataWidget`
to support multiple dataclass types.
"""

from .qt_compat import QtWidgets, QtCore
from .datawidget import DataWidget
from typing import Type, Optional
from .typing import DataclassProtocol


__all__ = [
    "DataclassStackedWidget",
    "DataclassTabWidget",
]


class DataclassStackedWidget(QtWidgets.QStackedWidget):
    """
    Stacked widget containing multiple :class:`DataWidget` and their dataclasses.

    To add :class:`DataWidget`, pass the widget and its dataclass to
    :meth:`addDataWidget` or to :meth:`insertDataWidget`.

    When the data value of current data widget changes, this widget emits
    :attr:`currentDataValueChanged` signal. When the current data widget is
    edited by user, :attr:`currentDataEdited` signal is emitted.

    """

    currentDataValueChanged = QtCore.Signal(dict)
    currentDataEdited = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self._dataWidgets = {}
        self._previousIndex = -1
        self.currentChanged.connect(self._onCurrentChange)

    def _onCurrentChange(self, index: int):
        """Handle the signals of old widget and current widget."""
        old = self.widget(self._previousIndex)
        if isinstance(old, DataWidget):
            old.dataValueChanged.disconnect(self.currentDataValueChanged)
            old.dataEdited.disconnect(self.currentDataEdited)
        new = self.widget(index)
        if isinstance(new, DataWidget):
            new.dataValueChanged.connect(self.currentDataValueChanged)
            new.dataEdited.connect(self.currentDataEdited)
        self._previousIndex = index

    def addDataWidget(
        self, widget: DataWidget, dataclass: Type[DataclassProtocol]
    ) -> int:
        """Add *widget* with binding it to *dataclass*."""
        index = self.addWidget(widget)
        self._dataWidgets[widget] = dataclass
        return index

    def insertDataWidget(
        self, index: int, widget: DataWidget, dataclass: Type[DataclassProtocol]
    ) -> int:
        index = self.insertWidget(index, widget)
        self._dataWidgets[widget] = dataclass
        return index

    def removeWidget(self, widget: QtWidgets.QWidget):
        if isinstance(widget, DataWidget):
            self._dataWidgets.pop(widget, None)
            if widget == self.currentWidget():
                widget.dataValueChanged.disconnect(self.currentDataValueChanged)
                widget.dataEdited.disconnect(self.currentDataEdited)
        super().removeWidget(widget)

    def currentDataclass(self) -> Optional[Type[DataclassProtocol]]:
        return self._dataWidgets.get(self.currentWidget())

    def indexOfDataclass(self, dataclass: Type[DataclassProtocol]) -> int:
        """Return the index of the widget bound to *dataclass*."""
        for widget, dcls in self._dataWidgets.items():
            if dcls == dataclass:
                index = self.indexOf(widget)
                break
        else:
            index = -1
        return index


class DataclassTabWidget(QtWidgets.QTabWidget):
    """
    Tab widget containing multiple :class:`DataWidget` and dataclasses.

    To add :class:`DataWidget`, pass the widget and its dataclass to
    :meth:`addDataWidget` or to :meth:`insertDataWidget`.

    When current index is changed by user, :attr:`activated` signal is emitted.
    When the data value of current data widget changes, this widget emits
    :attr:`currentDataValueChanged` signal. When the current data widget is
    edited by user, :attr:`currentDataEdited` signal is emitted.

    """

    activated = QtCore.Signal(int)
    currentDataValueChanged = QtCore.Signal(dict)
    currentDataEdited = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self._dataWidgets = {}
        self._previousIndex = -1
        self._blockActivated = False
        self.currentChanged.connect(self._onCurrentChange)

    def setCurrentIndex(self, index):
        self._blockActivated = True
        super().setCurrentIndex(index)
        self._blockActivated = False

    def _onCurrentChange(self, index: int):
        """Handle the signals of old widget and current widget."""
        old = self.widget(self._previousIndex)
        if isinstance(old, DataWidget):
            old.dataValueChanged.disconnect(self.currentDataValueChanged)
            old.dataEdited.disconnect(self.currentDataEdited)
        new = self.widget(index)
        if isinstance(new, DataWidget):
            new.dataValueChanged.connect(self.currentDataValueChanged)
            new.dataEdited.connect(self.currentDataEdited)
        self._previousIndex = index
        if not self._blockActivated:
            self.activated.emit(index)

    def addDataWidget(self, widget, dataclass, icon=None, label=None) -> int:
        """Add *widget* with binding it to *dataclass*."""
        args = [arg for arg in [icon, label] if arg is not None]
        index = self.addTab(widget, *args)
        self._dataWidgets[widget] = dataclass
        return index

    def insertDataWidget(self, index, widget, dataclass, icon=None, label=None) -> int:
        """Insert *widget* with binding it to *dataclass*."""
        args = [arg for arg in [icon, label] if arg is not None]
        index = self.insertTab(index, widget, *args)
        self._dataWidgets[widget] = dataclass
        return index

    def removeTab(self, index: int):
        widget = self.widget(index)
        if isinstance(widget, DataWidget):
            self._dataWidgets.pop(widget, None)
            if widget == self.currentWidget():
                widget.dataValueChanged.disconnect(self.currentDataValueChanged)
                widget.dataEdited.disconnect(self.currentDataEdited)
        super().removeTab(index)

    def currentDataclass(self) -> Optional[Type[DataclassProtocol]]:
        return self._dataWidgets.get(self.currentWidget())

    def indexOfDataclass(self, dataclass: Type[DataclassProtocol]) -> int:
        """Return the index of the widget bound to *dataclass*."""
        for widget, dcls in self._dataWidgets.items():
            if dcls == dataclass:
                index = self.indexOf(widget)
                break
        else:
            index = -1
        return index



import dataclasses
from dawiq import dataclass2Widget, DataclassStackedWidget, DataclassTabWidget
from dawiq.delegate import (
    convertFromQt,
    convertToQt,
    highlightEmptyField,
    DataclassDelegate,
    DataclassMapper,
)
from dawiq.qt_compat import QtGui, QtWidgets, QtCore
from typing import Tuple
import pytest


def test_convertFromQt():
    class CustomField:
        def __init__(self, x):
            self.x = x

        def __eq__(self, other):
            return type(self) == type(other) and self.x == other.x

    @dataclasses.dataclass
    class Cls0:
        a: CustomField = dataclasses.field(
            metadata=dict(fromQt_converter=lambda arg: CustomField(arg))
        )

    @dataclasses.dataclass
    class Cls1:
        x: int
        y: CustomField = dataclasses.field(
            metadata=dict(fromQt_converter=lambda arg: CustomField(arg))
        )
        z: Cls0

    assert convertFromQt(Cls1, dict(x=1, y=2, z=dict(a=3))) == dict(
        x=1, y=CustomField(2), z=dict(a=CustomField(3))
    )
    assert convertFromQt(Cls1, dict(x=None, y=None, z=None)) == dict()


def test_convertFromQt_defaultvalue():
    """Test that default value is ignored."""

    class CustomField:
        def __init__(self, x):
            self.x = x

        def __eq__(self, other):
            return type(self) == type(other) and self.x == other.x

    @dataclasses.dataclass
    class Cls0:
        x: CustomField = dataclasses.field(
            metadata=dict(fromQt_converter=lambda arg: CustomField(arg))
        )
        y: CustomField = dataclasses.field(
            default=CustomField(0),
            metadata=dict(fromQt_converter=lambda arg: CustomField(arg)),
        )
        z: int = 3

    assert convertFromQt(Cls0, dict(x=3, y=2, z=1)) == dict(
        x=CustomField(3), y=CustomField(2), z=1
    )
    assert convertFromQt(Cls0, dict()) == dict()
    assert convertFromQt(Cls0, dict(x=None, y=None, z=None)) == dict()

    @dataclasses.dataclass
    class Cls1:
        a: Cls0
        b: Cls0 = Cls0(x=CustomField(1))

    assert convertFromQt(
        Cls1, dict(a=dict(x=1, y=2, z=5), b=dict(x=3, y=2, z=1))
    ) == dict(
        a=dict(x=CustomField(1), y=CustomField(2), z=5),
        b=dict(x=CustomField(3), y=CustomField(2), z=1),
    )
    assert convertFromQt(Cls1, dict()) == dict()
    assert convertFromQt(Cls1, dict(a=None, b=None)) == dict()
    assert convertFromQt(Cls1, dict(a=None, b=dict(x=None, y=None, z=None))) == dict(
        b=dict()
    )

    @dataclasses.dataclass
    class Cls2:
        c: Cls1
        d: Cls1 = Cls1(Cls0(CustomField(10)))

    assert convertFromQt(Cls2, dict()) == dict()


def test_convertToQt():
    class CustomField:
        def __init__(self, x):
            self.x = x

        def __eq__(self, other):
            return type(self) == type(other) and self.x == other.x

    @dataclasses.dataclass
    class Cls0:
        a: CustomField = dataclasses.field(
            metadata=dict(toQt_converter=lambda val: val.x)
        )

    @dataclasses.dataclass
    class Cls1:
        x: int
        y: CustomField = dataclasses.field(
            metadata=dict(toQt_converter=lambda val: val.x)
        )
        z: Cls0

    assert convertToQt(
        Cls1, dict(x=1, y=CustomField(2), z=dict(a=CustomField(3)))
    ) == dict(x=1, y=2, z=dict(a=3))
    assert convertToQt(Cls1, dict()) == dict(x=None, y=None, z=None)


def test_convertToQt_defaultvalue():
    """Test that default value is ignored."""

    class CustomField:
        def __init__(self, x):
            self.x = x

        def __eq__(self, other):
            return type(self) == type(other) and self.x == other.x

    @dataclasses.dataclass
    class Cls0:
        x: CustomField = dataclasses.field(
            metadata=dict(toQt_converter=lambda val: val.x)
        )
        y: CustomField = dataclasses.field(
            default=CustomField(0),
            metadata=dict(toQt_converter=lambda val: val.x),
        )
        z: int = 3

    assert convertToQt(Cls0, dict()) == dict(x=None, y=None, z=None)

    @dataclasses.dataclass
    class Cls1:
        a: Cls0
        b: Cls0 = Cls0(x=CustomField(1))

    assert convertToQt(Cls1, dict()) == dict(a=None, b=None)
    assert convertToQt(Cls1, dict(b=dict())) == dict(
        a=None, b=dict(x=None, y=None, z=None)
    )

    @dataclasses.dataclass
    class Cls2:
        c: Cls1
        d: Cls1 = Cls1(Cls0(CustomField(10)))

    assert convertToQt(Cls2, dict()) == dict(c=None, d=None)
    assert convertToQt(Cls2, dict(d=dict())) == dict(c=None, d=dict(a=None, b=None))


def test_highlightEmptyField(qtbot):
    @dataclasses.dataclass
    class DataClass1:
        x: int

    editor1 = dataclass2Widget(DataClass1)

    highlightEmptyField(editor1, DataClass1)
    assert editor1.widget(0).property("requiresFieldValue")

    editor1.setDataValue(dict(x=10))
    highlightEmptyField(editor1, DataClass1)
    assert not editor1.widget(0).property("requiresFieldValue")

    @dataclasses.dataclass
    class DataClass2:
        x: int = 1

    editor2 = dataclass2Widget(DataClass2)

    highlightEmptyField(editor2, DataClass2)
    assert not editor2.widget(0).property("requiresFieldValue")

    editor2.setDataValue(dict(x=10))
    highlightEmptyField(editor2, DataClass2)
    assert not editor2.widget(0).property("requiresFieldValue")


def test_highlightEmptyField_recursive(qtbot):
    @dataclasses.dataclass
    class DataClass1:
        x: int
        y: int = 3

    @dataclasses.dataclass
    class DataClass2:
        a: DataClass1
        b: DataClass1 = DataClass1(1, 2)

    editor = dataclass2Widget(DataClass2)
    highlightEmptyField(editor, DataClass2)
    assert editor.widget(0).widget(0).property("requiresFieldValue")
    assert not editor.widget(0).widget(1).property("requiresFieldValue")
    assert not editor.widget(1).widget(0).property("requiresFieldValue")
    assert not editor.widget(1).widget(1).property("requiresFieldValue")


def test_highlightEmptyField_noDataclass(qtbot):
    @dataclasses.dataclass
    class DataClass1:
        x: int
        y: int

    @dataclasses.dataclass
    class DataClass2:
        a: DataClass1

    editor = dataclass2Widget(DataClass2)
    highlightEmptyField(editor, None)
    assert not editor.widget(0).widget(0).property("requiresFieldValue")
    assert not editor.widget(0).widget(1).property("requiresFieldValue")


@dataclasses.dataclass
class DataClass1:
    x: bool


@dataclasses.dataclass
class DataClass2:
    a: bool
    b: bool
    c: bool
    d: bool
    e: bool


@dataclasses.dataclass
class DataClass3:
    x: bool


@pytest.fixture
def dataclassStackedWidget(qtbot):
    widget = DataclassStackedWidget()
    for dcls in [DataClass1, DataClass2]:
        widget.addDataWidget(dataclass2Widget(dcls), dcls)

    return widget


@pytest.fixture
def dataclassTabWidget(qtbot):
    widget = DataclassTabWidget()
    for dcls in [DataClass1, DataClass2]:
        widget.addDataWidget(dataclass2Widget(dcls), dcls, dcls.__name__)

    return widget


# test DataclassDelegate


def test_DataclassDelegate_setModelData(qtbot):
    @dataclasses.dataclass
    class Dcls:
        x: int

    delegate = DataclassDelegate()
    model = QtGui.QStandardItemModel()

    item = QtGui.QStandardItem()
    item.setData(Dcls, role=DataclassDelegate.TypeRole)
    model.appendRow(item)

    dataWidget = dataclass2Widget(Dcls)
    mapper = QtWidgets.QDataWidgetMapper()

    mapper.setModel(model)
    mapper.addMapping(dataWidget, 0)
    mapper.setItemDelegate(delegate)

    modelIndex = model.index(0, 0)
    mapper.setCurrentModelIndex(modelIndex)
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) is None

    mapper.submit()
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) == dict()

    dataWidget.widget(0).setText("0")
    mapper.submit()
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) == dict(x=0)


def test_DataclassDelegate_setModelData_dataclassStackedWidget(
    qtbot, dataclassStackedWidget
):
    delegate = DataclassDelegate()
    model = QtGui.QStandardItemModel()

    item = QtGui.QStandardItem()
    item.setData(DataClass1, role=DataclassDelegate.TypeRole)
    model.appendRow(item)

    mapper = QtWidgets.QDataWidgetMapper()
    mapper.setModel(model)
    mapper.addMapping(dataclassStackedWidget, 0)
    mapper.setItemDelegate(delegate)

    modelIndex = model.index(0, 0)
    mapper.setCurrentModelIndex(modelIndex)
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) is None

    mapper.submit()
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) == dict(x=False)

    dataclassStackedWidget.currentWidget().widget(0).click()
    mapper.submit()
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) == dict(x=True)


def test_DataclassDelegate_setModelData_dataclassTabWidget(qtbot, dataclassTabWidget):
    delegate = DataclassDelegate()
    model = QtGui.QStandardItemModel()

    item = QtGui.QStandardItem()
    item.setData(DataClass1, role=DataclassDelegate.TypeRole)
    model.appendRow(item)

    mapper = QtWidgets.QDataWidgetMapper()
    mapper.setModel(model)
    mapper.addMapping(dataclassTabWidget, 0)
    mapper.setItemDelegate(delegate)

    modelIndex = model.index(0, 0)
    mapper.setCurrentModelIndex(modelIndex)
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) is None

    mapper.submit()
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) == dict(x=False)

    dataclassTabWidget.currentWidget().widget(0).click()
    mapper.submit()
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) == dict(x=True)


def test_DataclassDelegate_setEditorData(qtbot):
    @dataclasses.dataclass
    class Dcls:
        x: int

    delegate = DataclassDelegate()
    model = QtGui.QStandardItemModel()

    for i in range(3):
        item = QtGui.QStandardItem()
        item.setData(Dcls, role=DataclassDelegate.TypeRole)
        model.appendRow(item)

    dataWidget = dataclass2Widget(Dcls)
    mapper = QtWidgets.QDataWidgetMapper()

    mapper.setModel(model)
    mapper.addMapping(dataWidget, 0)
    mapper.setItemDelegate(delegate)

    modelIndex0 = model.index(0, 0)
    model.setData(modelIndex0, dict(x=0), role=DataclassDelegate.DataRole)
    modelIndex1 = model.index(1, 0)
    model.setData(modelIndex1, dict(x=1), role=DataclassDelegate.DataRole)
    modelIndex2 = model.index(2, 0)
    model.setData(modelIndex2, dict(), role=DataclassDelegate.DataRole)

    assert dataWidget.dataValue() == dict(x=None)
    assert dataWidget.widget(0).text() == ""

    mapper.setCurrentModelIndex(modelIndex0)
    assert dataWidget.dataValue() == dict(x=0)
    assert dataWidget.widget(0).text() == "0"

    mapper.setCurrentModelIndex(modelIndex1)
    assert dataWidget.dataValue() == dict(x=1)
    assert dataWidget.widget(0).text() == "1"

    mapper.setCurrentModelIndex(modelIndex2)
    assert dataWidget.dataValue() == dict(x=None)
    assert dataWidget.widget(0).text() == ""

    model.setData(modelIndex2, dict(x=10), role=DataclassDelegate.DataRole)
    assert dataWidget.dataValue() == dict(x=10)
    assert dataWidget.widget(0).text() == "10"


# test DataclassMapper


def test_DataclassMapper_addMapping_dataWidget(qtbot):
    @dataclasses.dataclass
    class Dcls:
        x: int
        y: bool

    delegate = DataclassDelegate()
    model = QtGui.QStandardItemModel()

    item = QtGui.QStandardItem()
    item.setData(Dcls, role=DataclassDelegate.TypeRole)
    model.appendRow(item)

    dataWidget = dataclass2Widget(Dcls)
    mapper = DataclassMapper()

    mapper.setModel(model)
    mapper.addMapping(dataWidget, 0)
    mapper.setItemDelegate(delegate)

    modelIndex = model.index(0, 0)
    mapper.setCurrentModelIndex(modelIndex)
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) is None  # IMPORTANT!

    dataWidget.widget(0).setText("0")
    qtbot.keyPress(dataWidget.widget(0), QtCore.Qt.Key.Key_Return)
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) == dict(x=0, y=False)

    dataWidget.widget(1).click()
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) == dict(x=0, y=True)


def test_DataclassMapper_addMapping_dataclassStackedWidget(
    qtbot, dataclassStackedWidget
):
    model = QtGui.QStandardItemModel()
    item = QtGui.QStandardItem()
    item.setData(DataClass1, role=DataclassDelegate.TypeRole)
    model.appendRow(item)

    delegate = DataclassDelegate()
    mapper = DataclassMapper()
    mapper.setItemDelegate(delegate)
    mapper.setModel(model)
    mapper.addMapping(dataclassStackedWidget, 0)

    modelIndex = model.index(0, 0)
    mapper.setCurrentModelIndex(modelIndex)
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) is None

    dataclassStackedWidget.currentWidget().widget(0).click()
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) == dict(x=True)

    # data type change
    item.setData(DataClass2, role=DataclassDelegate.TypeRole)
    assert (
        dataclassStackedWidget.currentIndex()
        == dataclassStackedWidget.indexOfDataclass(DataClass2)
    )
    item.setData(DataClass1, role=DataclassDelegate.TypeRole)
    assert (
        dataclassStackedWidget.currentIndex()
        == dataclassStackedWidget.indexOfDataclass(DataClass1)
    )


def test_DataclassMapper_addMapping_dataclassTabWidget(qtbot, dataclassTabWidget):
    model = QtGui.QStandardItemModel()
    item = QtGui.QStandardItem()
    item.setData(DataClass1, role=DataclassDelegate.TypeRole)
    model.appendRow(item)

    delegate = DataclassDelegate()
    mapper = DataclassMapper()
    mapper.setItemDelegate(delegate)
    mapper.setModel(model)
    mapper.addMapping(dataclassTabWidget, 0)

    modelIndex = model.index(0, 0)
    mapper.setCurrentModelIndex(modelIndex)
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) is None

    dataclassTabWidget.currentWidget().widget(0).click()
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) == dict(x=True)

    # data type change
    item.setData(DataClass2, role=DataclassDelegate.TypeRole)
    assert dataclassTabWidget.currentIndex() == dataclassTabWidget.indexOfDataclass(
        DataClass2
    )
    item.setData(DataClass1, role=DataclassDelegate.TypeRole)
    assert dataclassTabWidget.currentIndex() == dataclassTabWidget.indexOfDataclass(
        DataClass1
    )


def test_DataclassMapper_removeMapping_dataWidget(qtbot):
    @dataclasses.dataclass
    class Dcls:
        x: int
        y: bool

    delegate = DataclassDelegate()
    model = QtGui.QStandardItemModel()

    item = QtGui.QStandardItem()
    item.setData(Dcls, role=DataclassDelegate.TypeRole)
    model.appendRow(item)

    dataWidget = dataclass2Widget(Dcls)
    mapper = DataclassMapper()

    mapper.setModel(model)
    mapper.addMapping(dataWidget, 0)
    mapper.setItemDelegate(delegate)

    mapper.removeMapping(dataWidget)

    modelIndex = model.index(0, 0)
    mapper.setCurrentModelIndex(modelIndex)
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) is None

    dataWidget.widget(0).setText("0")
    qtbot.keyPress(dataWidget.widget(0), QtCore.Qt.Key.Key_Return)
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) is None

    dataWidget.widget(1).click()
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) is None


def test_DataclassMapper_removeMapping_dataclassStackedWidget(
    qtbot, dataclassStackedWidget
):
    model = QtGui.QStandardItemModel()
    item = QtGui.QStandardItem()
    item.setData(DataClass1, role=DataclassDelegate.TypeRole)
    model.appendRow(item)

    delegate = DataclassDelegate()
    mapper = DataclassMapper()
    mapper.setItemDelegate(delegate)
    mapper.setModel(model)
    mapper.addMapping(dataclassStackedWidget, 0)

    modelIndex = model.index(0, 0)
    mapper.setCurrentModelIndex(modelIndex)

    mapper.removeMapping(dataclassStackedWidget)

    dataclassStackedWidget.currentWidget().widget(0).click()
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) is None

    # data type change
    dataclassStackedWidget.setCurrentIndex(0)
    assert model.data(modelIndex, role=DataclassDelegate.TypeRole) == DataClass1
    dataclassStackedWidget.setCurrentIndex(2)
    assert model.data(modelIndex, role=DataclassDelegate.TypeRole) == DataClass1


def test_DataclassMapper_removeMapping_dataclassTabWidget(qtbot, dataclassTabWidget):
    model = QtGui.QStandardItemModel()
    item = QtGui.QStandardItem()
    item.setData(DataClass1, role=DataclassDelegate.TypeRole)
    model.appendRow(item)

    delegate = DataclassDelegate()
    mapper = DataclassMapper()
    mapper.setItemDelegate(delegate)
    mapper.setModel(model)
    mapper.addMapping(dataclassTabWidget, 0)

    modelIndex = model.index(0, 0)
    mapper.setCurrentModelIndex(modelIndex)

    mapper.removeMapping(dataclassTabWidget)

    dataclassTabWidget.currentWidget().widget(0).click()
    assert model.data(modelIndex, role=DataclassDelegate.DataRole) is None

    # data type change
    dataclassTabWidget.setCurrentIndex(0)
    assert model.data(modelIndex, role=DataclassDelegate.TypeRole) == DataClass1
    dataclassTabWidget.setCurrentIndex(2)
    assert model.data(modelIndex, role=DataclassDelegate.TypeRole) == DataClass1


def test_DataclassMapper_Tuple_setCurrentIndex_crash(qtbot):
    """Test that setting index to nested widget does not cause infinite loop."""

    @dataclasses.dataclass
    class DataClass:
        x: Tuple[int]

    delegate = DataclassDelegate()
    model = QtGui.QStandardItemModel()

    item = QtGui.QStandardItem()
    item.setData(DataClass, role=DataclassDelegate.TypeRole)
    model.appendRow(item)

    dataWidget = dataclass2Widget(DataClass)
    mapper = DataclassMapper()

    mapper.setModel(model)
    mapper.addMapping(dataWidget, 0)
    mapper.setItemDelegate(delegate)

    mapper.setCurrentIndex(0)  # must not crash


def test_DataclassMapper_default(qtbot):
    """Test that default value of dataclass is not applied to widget & model."""

    @dataclasses.dataclass
    class DataClass:
        x: int = 3

    delegate = DataclassDelegate()
    model = QtGui.QStandardItemModel()

    item = QtGui.QStandardItem()
    item.setData(DataClass, role=DataclassDelegate.TypeRole)
    model.appendRow(item)

    dataWidget = dataclass2Widget(DataClass)
    mapper = DataclassMapper()

    mapper.setModel(model)
    mapper.addMapping(dataWidget, 0)
    mapper.setItemDelegate(delegate)

    modelIndex = model.index(0, 0)
    mapper.setCurrentModelIndex(modelIndex)

    assert model.data(modelIndex, role=DataclassDelegate.DataRole) is None
    assert dataWidget.dataValue() == dict(x=None)



from dawiq import (
    BoolCheckBox,
    EmptyIntValidator,
    IntLineEdit,
    EmptyFloatValidator,
    FloatLineEdit,
    StrLineEdit,
    EnumComboBox,
    TupleGroupBox,
)
import enum
from dawiq.qt_compat import QtCore, QtWidgets


def test_BoolCheckBox(qtbot):
    widget = BoolCheckBox()
    widget.setCheckState(QtCore.Qt.CheckState.Unchecked)

    # test value change by setFieldValue
    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val is True,
    ):
        widget.setFieldValue(True)
    assert widget.fieldValue() is True
    assert widget.checkState() == QtCore.Qt.CheckState.Checked

    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val is False,
    ):
        widget.setFieldValue(False)
    assert widget.fieldValue() is False
    assert widget.checkState() == QtCore.Qt.CheckState.Unchecked

    widget.setCheckState(QtCore.Qt.CheckState.Unchecked)

    # test value change by clicking
    with qtbot.waitSignals(
        [widget.fieldValueChanged, widget.fieldEdited],
        check_params_cbs=[lambda val: val is True, lambda: True],
    ):
        widget.click()
    assert widget.fieldValue() is True
    assert widget.checkState() == QtCore.Qt.CheckState.Checked

    with qtbot.waitSignals(
        [widget.fieldValueChanged, widget.fieldEdited],
        check_params_cbs=[lambda val: val is False, lambda: True],
    ):
        widget.click()
    assert widget.fieldValue() is False
    assert widget.checkState() == QtCore.Qt.CheckState.Unchecked


def test_BoolCheckBox_tristate(qtbot):
    widget = BoolCheckBox()
    widget.setCheckState(QtCore.Qt.CheckState.Checked)

    widget.setTristate(False)
    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val is False,
    ):
        widget.setFieldValue(None)
    assert widget.fieldValue() is False
    assert widget.checkState() == QtCore.Qt.CheckState.Unchecked

    widget.setTristate(True)
    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val is None,
    ):
        widget.setFieldValue(None)
    assert widget.fieldValue() is None
    assert widget.checkState() == QtCore.Qt.CheckState.PartiallyChecked


def test_BoolCheckBox_setRequired(qtbot):
    """Check box always has field value."""
    widget = BoolCheckBox()

    widget.setRequired(True)
    assert not widget.property("requiresFieldValue")
    widget.click()
    widget.setRequired(True)
    assert not widget.property("requiresFieldValue")

    widget.setRequired(False)
    assert not widget.property("requiresFieldValue")
    widget.click()
    widget.setRequired(False)
    assert not widget.property("requiresFieldValue")


def test_EmptyIntValidator(qtbot):
    widget = QtWidgets.QLineEdit()
    widget.setValidator(EmptyIntValidator(widget))

    # empty text is valid
    with qtbot.waitSignal(widget.editingFinished):
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)

    with qtbot.assertNotEmitted(widget.editingFinished):
        qtbot.keyPress(widget, "-")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)

    with qtbot.waitSignal(widget.editingFinished):
        qtbot.keyPress(widget, "1")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)


def test_IntLineEdit(qtbot):
    widget = IntLineEdit()

    assert widget.fieldValue() is None
    assert not widget.text()

    # test value change by setFieldValue
    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val == 1,
    ):
        widget.setFieldValue(1)
    assert widget.fieldValue() == 1
    assert widget.text() == "1"

    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val is None,
    ):
        widget.setFieldValue(None)
    assert widget.fieldValue() is None
    assert not widget.text()

    # test value change by keyboard
    widget.clear()
    with qtbot.waitSignals(
        [widget.fieldEdited],
        check_params_cbs=[lambda: True],
    ):
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)
    assert widget.fieldValue() is None

    with qtbot.waitSignals(
        [widget.fieldValueChanged, widget.fieldEdited],
        check_params_cbs=[lambda val: val == 1, lambda: True],
    ):
        qtbot.keyPress(widget, "1")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)
    assert widget.fieldValue() == 1


def test_IntLineEdit_setRequired(qtbot):
    widget = IntLineEdit()

    widget.setRequired(True)
    assert widget.property("requiresFieldValue")
    widget.setFieldValue(10)
    widget.setRequired(True)
    assert not widget.property("requiresFieldValue")

    widget.clear()
    widget.setRequired(False)
    assert not widget.property("requiresFieldValue")
    widget.setRequired(True)
    assert widget.property("requiresFieldValue")


def test_EmptyFloatValidator(qtbot):
    widget = QtWidgets.QLineEdit()
    widget.setValidator(EmptyFloatValidator(widget))

    # empty text is valid
    with qtbot.waitSignal(widget.editingFinished):
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)

    with qtbot.assertNotEmitted(widget.editingFinished):
        qtbot.keyPress(widget, "-")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)

    with qtbot.waitSignal(widget.editingFinished):
        qtbot.keyPress(widget, "1")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)

    with qtbot.waitSignal(widget.editingFinished):
        qtbot.keyPress(widget, ".")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)

    with qtbot.waitSignal(widget.editingFinished):
        qtbot.keyPress(widget, "1")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)


def test_FloatLineEdit(qtbot):
    widget = FloatLineEdit()

    assert widget.fieldValue() is None
    assert not widget.text()

    # test value change by setFieldValue
    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val == 1.2,
    ):
        widget.setFieldValue(1.2)
    assert widget.fieldValue() == 1.2
    assert widget.text() == "1.2"

    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val is None,
    ):
        widget.setFieldValue(None)
    assert widget.fieldValue() is None
    assert not widget.text()

    # test value change by keyboard
    widget.clear()
    with qtbot.waitSignals(
        [widget.fieldEdited],
        check_params_cbs=[lambda: True],
    ):
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)
    assert widget.fieldValue() is None

    with qtbot.waitSignals(
        [widget.fieldValueChanged, widget.fieldEdited],
        check_params_cbs=[lambda val: val == 1.2, lambda: True],
    ):
        qtbot.keyPress(widget, "1")
        qtbot.keyPress(widget, ".")
        qtbot.keyPress(widget, "2")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)
    assert widget.fieldValue() == 1.2


def test_FloatLineEdit_setRequired(qtbot):
    widget = FloatLineEdit()

    widget.setRequired(True)
    assert widget.property("requiresFieldValue")
    widget.setFieldValue(10.0)
    widget.setRequired(True)
    assert not widget.property("requiresFieldValue")

    widget.clear()
    widget.setRequired(False)
    assert not widget.property("requiresFieldValue")
    widget.setRequired(True)
    assert widget.property("requiresFieldValue")


def test_StrLineEdit(qtbot):
    widget = StrLineEdit()

    assert widget.fieldValue() == ""
    assert not widget.text()

    # test value change by setFieldValue
    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val == "1",
    ):
        widget.setFieldValue("1")
    assert widget.fieldValue() == "1"
    assert widget.text() == "1"
    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val == "x",
    ):
        widget.setFieldValue("x")
    assert widget.fieldValue() == "x"
    assert widget.text() == "x"

    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val == "",
    ):
        widget.setFieldValue(None)
    assert widget.fieldValue() == ""
    assert widget.text() == ""

    # test value change by keyboard
    widget.clear()
    with qtbot.waitSignals(
        [widget.fieldEdited],
        check_params_cbs=[lambda: True],
    ):
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)
    assert widget.fieldValue() == ""

    with qtbot.waitSignals(
        [widget.fieldValueChanged, widget.fieldEdited],
        check_params_cbs=[lambda val: val == "x", lambda: True],
    ):
        qtbot.keyPress(widget, "x")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)
    assert widget.fieldValue() == "x"


def test_StrLineEdit_setRequired(qtbot):
    """String line edit always has field value."""
    widget = StrLineEdit()

    widget.setRequired(True)
    assert not widget.property("requiresFieldValue")
    widget.setFieldValue("spam")
    widget.setRequired(True)
    assert not widget.property("requiresFieldValue")

    widget.clear()
    widget.setRequired(False)
    assert not widget.property("requiresFieldValue")
    widget.setRequired(True)
    assert not widget.property("requiresFieldValue")


def test_EnumComboBox(qtbot):
    class MyEnum(enum.Enum):
        x = 1
        y = 2
        z = 3

    widget = EnumComboBox.fromEnum(MyEnum)

    assert widget.count() == 3
    assert widget.currentIndex() == -1
    assert widget.fieldValue() is None

    # test with setFieldValue
    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val == MyEnum.x,
    ):
        widget.setFieldValue(MyEnum.x)
    assert widget.currentIndex() == 0
    assert widget.fieldValue() == MyEnum.x

    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val == MyEnum.y,
    ):
        widget.setFieldValue(MyEnum.y)
    assert widget.currentIndex() == 1
    assert widget.fieldValue() == MyEnum.y

    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val == MyEnum.z,
    ):
        widget.setFieldValue(MyEnum.z)
    assert widget.currentIndex() == 2
    assert widget.fieldValue() == MyEnum.z

    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val is None,
    ):
        widget.setFieldValue(None)
    assert widget.currentIndex() == -1
    assert widget.fieldValue() is None

    # test with setCurrentIndex
    with qtbot.waitSignals(
        [widget.fieldValueChanged, widget.fieldEdited],
        check_params_cbs=[lambda val: val == MyEnum.x, lambda: True],
    ):
        qtbot.keyClick(widget, MyEnum.x.name)
    assert widget.currentIndex() == 0
    assert widget.fieldValue() == MyEnum.x


def test_EnumComboBox_setRequired(qtbot):
    class MyEnum(enum.Enum):
        x = 1
        y = 2
        z = 3

    widget = EnumComboBox.fromEnum(MyEnum)

    widget.setRequired(True)
    assert widget.property("requiresFieldValue")
    widget.setCurrentIndex(1)
    widget.setRequired(True)
    assert not widget.property("requiresFieldValue")

    widget.setRequired(False)
    assert not widget.property("requiresFieldValue")
    widget.setCurrentIndex(-1)
    widget.setRequired(False)
    assert not widget.property("requiresFieldValue")


def test_TupleGroupBox_addWidget(qtbot):
    widget = TupleGroupBox()
    assert widget.count() == 0

    w0 = BoolCheckBox()
    widget.addWidget(w0)
    assert widget.count() == 1
    assert widget.widget(0) is w0
    assert widget.widget(1) is None

    w1 = IntLineEdit()
    widget.addWidget(w1)
    assert widget.count() == 2
    assert widget.widget(0) is w0
    assert widget.widget(1) is w1

    # test that signals are connected
    with qtbot.waitSignals([widget.fieldValueChanged, widget.fieldEdited]):
        w0.click()
    with qtbot.waitSignals([widget.fieldValueChanged, widget.fieldEdited]):
        qtbot.keyPress(w1, "1")
        qtbot.keyPress(w1, QtCore.Qt.Key.Key_Return)


def test_TupleGroupBox_insertWidget(qtbot):
    widget = TupleGroupBox()
    assert widget.count() == 0

    w0 = BoolCheckBox()
    widget.insertWidget(0, w0)
    assert widget.count() == 1
    assert widget.widget(0) is w0
    assert widget.widget(1) is None

    w1 = IntLineEdit()
    widget.insertWidget(0, w1)
    assert widget.count() == 2
    assert widget.widget(0) is w1
    assert widget.widget(1) is w0

    # test that signals are connected
    with qtbot.waitSignals([widget.fieldValueChanged, widget.fieldEdited]):
        w0.click()
    with qtbot.waitSignals([widget.fieldValueChanged, widget.fieldEdited]):
        qtbot.keyPress(w1, "1")
        qtbot.keyPress(w1, QtCore.Qt.Key.Key_Return)


def test_TupleGroupBox_removeWidget(qtbot):
    widget = TupleGroupBox()
    w0 = BoolCheckBox()
    w1 = IntLineEdit()

    widget.addWidget(w0)
    assert widget.count() == 1

    widget.removeWidget(w1)
    assert widget.count() == 1

    widget.removeWidget(w0)
    assert widget.count() == 0

    # test that signals are disconnected
    with qtbot.assertNotEmitted(widget.fieldValueChanged):
        w0.click()
    with qtbot.assertNotEmitted(widget.fieldEdited):
        w0.click()
    with qtbot.assertNotEmitted(widget.fieldValueChanged):
        qtbot.keyPress(w1, "1")
        qtbot.keyPress(w1, QtCore.Qt.Key.Key_Return)
    with qtbot.assertNotEmitted(widget.fieldEdited):
        qtbot.keyPress(w1, "1")
        qtbot.keyPress(w1, QtCore.Qt.Key.Key_Return)


def test_TupleGroupBox_fieldValue(qtbot):
    widget = TupleGroupBox()
    widget.addWidget(IntLineEdit())
    widget.addWidget(IntLineEdit())
    assert widget.fieldValue() == (None, None)

    widget.widget(0).setText("1")
    widget.widget(1).setText("2")
    assert widget.fieldValue() == (1, 2)


def test_TupleGroupBox_setFieldValue(qtbot):
    widget = TupleGroupBox()
    widget.addWidget(IntLineEdit())
    widget.addWidget(IntLineEdit())

    class Counter:
        def __init__(self):
            self.i = 0

        def count(self):
            self.i += 1

        def reset(self):
            self.i = 0

    counter = Counter()
    widget.fieldValueChanged.connect(counter.count)

    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val == (1, 2),
    ):
        widget.setFieldValue((1, 2))
    assert widget.fieldValue() == (1, 2)
    assert counter.i == 1

    counter.reset()
    with qtbot.waitSignal(
        widget.fieldValueChanged,
        check_params_cb=lambda val: val == (None, None),
    ):
        widget.setFieldValue(None)
    assert widget.fieldValue() == (None, None)
    assert counter.i == 1


def test_TupleGroupBox_subwidget(qtbot):
    widget = TupleGroupBox()
    widget.addWidget(IntLineEdit())
    widget.addWidget(IntLineEdit())

    with qtbot.waitSignals(
        [widget.fieldValueChanged, widget.fieldEdited],
        check_params_cbs=[lambda val: val == (1, None), lambda: True],
    ):
        qtbot.keyPress(widget.widget(0), "1")
        qtbot.keyPress(widget.widget(0), QtCore.Qt.Key.Key_Return)
    assert widget.fieldValue() == (1, None)

    with qtbot.waitSignals(
        [widget.fieldValueChanged, widget.fieldEdited],
        check_params_cbs=[lambda val: val == (1, 2), lambda: True],
    ):
        qtbot.keyPress(widget.widget(1), "2")
        qtbot.keyPress(widget.widget(1), QtCore.Qt.Key.Key_Return)
    assert widget.fieldValue() == (1, 2)


def test_TupleGroupBox_setRequired(qtbot):
    widget = TupleGroupBox()
    widget.addWidget(IntLineEdit())
    widget.addWidget(IntLineEdit())

    widget.setRequired(True)
    assert widget.widget(0).property("requiresFieldValue")
    assert widget.widget(1).property("requiresFieldValue")
    widget.setFieldValue((None, 1))
    widget.setRequired(True)
    assert widget.widget(0).property("requiresFieldValue")
    assert not widget.widget(1).property("requiresFieldValue")
    widget.setFieldValue((0, 1))
    widget.setRequired(True)
    assert not widget.widget(0).property("requiresFieldValue")
    assert not widget.widget(1).property("requiresFieldValue")

    widget.widget(0).clear()
    widget.widget(1).clear()

    widget.setRequired(False)
    assert not widget.widget(0).property("requiresFieldValue")
    assert not widget.widget(1).property("requiresFieldValue")
    widget.setFieldValue((None, 1))
    widget.setRequired(False)
    assert not widget.widget(0).property("requiresFieldValue")
    assert not widget.widget(1).property("requiresFieldValue")
    widget.setFieldValue((0, 1))
    widget.setRequired(False)
    assert not widget.widget(0).property("requiresFieldValue")
    assert not widget.widget(1).property("requiresFieldValue")



"""
Data widget
===========

:mod:`dawiq.datawidget` provides :class:`DataWidget` to represent the data
structure established by the dataclass.
"""

from .qt_compat import QtCore, QtWidgets
from .fieldwidgets import (
    BoolCheckBox,
    IntLineEdit,
    FloatLineEdit,
    StrLineEdit,
    EnumComboBox,
    TupleGroupBox,
)
import dataclasses
from enum import Enum
from typing import Optional, Any, Union, Type, Callable, Dict, get_type_hints
from .typing import FieldWidgetProtocol, DataclassProtocol


__all__ = [
    "DataWidget",
    "type2Widget",
    "dataclass2Widget",
]


class DataWidget(QtWidgets.QGroupBox):
    """
    Group box for structured data.

    This group box contains the field widgets as subwidgets. :meth:`dataValue`
    returns the dictionary of field values and :meth:`setDataValue` sets the
    values to the subwidgets.

    Whenever the data value changes, :attr:`dataValueChanged` signal is emitted.
    When any field is edited by the user, :attr:`dataEdited` signal is emitted.

    This widget is also used to represent the nested dataclass field, therefore
    it follows :class:`FieldWidgetProtocol`.

    Notes
    =====

    This class can be constructed from :func:`dataclass2Widget`, but the widget
    does not store the dataclass type. To associate the data widget to the
    dataclass use :class:`DataclassDelegate`.

    """

    dataValueChanged = QtCore.Signal(dict)
    dataEdited = QtCore.Signal()
    fieldValueChanged = dataValueChanged
    fieldEdited = dataEdited

    def __init__(
        self,
        orientation: QtCore.Qt.Orientation = QtCore.Qt.Orientation.Vertical,
        parent=None,
    ):
        super().__init__(parent)
        self._orientation = orientation

        if orientation == QtCore.Qt.Orientation.Vertical:
            layout = QtWidgets.QVBoxLayout()
        elif orientation == QtCore.Qt.Orientation.Horizontal:
            layout = QtWidgets.QHBoxLayout()
        else:
            raise TypeError(f"Invalid orientation: {orientation}")
        self.setLayout(layout)

    def orientation(self) -> QtCore.Qt.Orientation:
        """Orientation to stack the subwidgets."""
        return self._orientation

    def count(self) -> int:
        """Number of subwidgets."""
        return self.layout().count()

    def widget(self, index: int) -> Optional[FieldWidgetProtocol]:
        """
        Returns the subwidget at the given index, or None for invalid index.
        """
        item = self.layout().itemAt(index)
        if item is not None:
            item = item.widget()
        return item

    def insertWidget(
        self,
        index: int,
        widget: FieldWidgetProtocol,
        stretch: int = 0,
        alignment: QtCore.Qt.AlignmentFlag = QtCore.Qt.AlignmentFlag(0),
    ):
        """Insert the widget to layout and connect the signals."""
        for i in range(self.count()):
            w = self.widget(i)
            if w is None:
                break
            if widget.fieldName() == w.fieldName():
                raise KeyError(f"Data name '{widget.fieldName()}' is duplicate")
        widget.fieldValueChanged.connect(self._onSubfieldValueChange)
        widget.fieldEdited.connect(self.dataEdited)
        self.layout().insertWidget(index, widget, stretch, alignment)

    def addWidget(
        self,
        widget: FieldWidgetProtocol,
        stretch: int = 0,
        alignment: QtCore.Qt.AlignmentFlag = QtCore.Qt.AlignmentFlag(0),
    ):
        """Add the widget to layout and connect the signals."""
        for i in range(self.count()):
            w = self.widget(i)
            if w is None:
                break
            if widget.fieldName() == w.fieldName():
                raise KeyError(f"Data name '{widget.fieldName()}' is duplicate")
        widget.fieldValueChanged.connect(self._onSubfieldValueChange)
        widget.fieldEdited.connect(self.dataEdited)
        self.layout().addWidget(widget, stretch, alignment)

    def removeWidget(self, widget: FieldWidgetProtocol):
        """Remove the widget from layout and disconnect the signals."""
        for i in range(self.count()):
            w = self.widget(i)
            if w is None:
                break
            if w == widget:
                widget.fieldValueChanged.disconnect(self._onSubfieldValueChange)
                widget.fieldEdited.disconnect(self.dataEdited)
                break
        self.layout().removeWidget(widget)

    def dataValue(self) -> Dict[str, Any]:
        ret = {}
        for i in range(self.count()):
            w = self.widget(i)
            if w is None:
                break
            ret[w.fieldName()] = w.fieldValue()
        return ret

    fieldValue = dataValue

    def setDataValue(self, data: Optional[Dict[str, Any]]):
        if data is None:
            data = {}

        for i in range(self.count()):
            w = self.widget(i)
            if w is None:
                break
            val = data.get(w.fieldName(), None)  # type: ignore[union-attr]
            w.fieldValueChanged.disconnect(self._onSubfieldValueChange)
            try:
                w.setFieldValue(val)
            except TypeError:
                w.setFieldValue(None)
            w.fieldValueChanged.connect(self._onSubfieldValueChange)
        self.dataValueChanged.emit(data)

    setFieldValue = setDataValue

    def _onSubfieldValueChange(self, value: Any):
        self.dataValueChanged.emit(self.dataValue())

    def fieldName(self) -> str:
        return self.title()

    def setFieldName(self, name: str):
        self.setTitle(name)

    def setRequired(self, required: bool):
        """Recursively set *required* to all subwidgets."""
        for i in range(self.count()):
            widget = self.widget(i)
            if widget is None:
                continue
            widget.setRequired(required)


def type2Widget(t: Any) -> FieldWidgetProtocol:
    """
    Construct the widget for given type annotation *t*.

    The following types are supported. Dataclass type is not converted here but
    by :func:`dataclass2Widget`.

    * :class:`enum.Enum` -> :class:`.EnumComboBox`
    * :class:`bool` -> :class:`.BoolCheckBox`
    * :obj:`Optional[bool]` -> :class:`.BoolCheckBox` with tristate
    * :class:`int` or :obj:`Optional[int]` -> :class:`.IntLineEdit`
    * :class:`float` or :obj:`Optional[float]` -> :class:`.FloatLineEdit`
    * :class:`str` or :obj:`Optional[str]` -> :class:`.StrLineEdit`
    * :obj:`Tuple` -> :class:`.TupleGroupBox` with nested field widgets

    For :obj:`Tuple`, its length must be finite (no :class:`Ellipsis` in args)
    and item types must be the supported type.

    """
    # When new type is supported, update intro.rst as well

    if isinstance(t, type) and issubclass(t, Enum):
        return EnumComboBox.fromEnum(t)
    if t is bool:
        return BoolCheckBox()
    if t is int:
        return IntLineEdit()
    if t is float:
        return FloatLineEdit()
    if t is str:
        return StrLineEdit()

    origin = getattr(t, "__origin__", None)  # t is tuple

    if origin is tuple:
        args = getattr(t, "__args__", None)
        if args is None:
            raise TypeError("%s does not have argument type" % t)
        if Ellipsis in args:
            txt = "Number of arguments of %s not fixed" % t
            raise TypeError(txt)

        subwidgets = [type2Widget(arg) for arg in args]
        tupwidget = TupleGroupBox()
        for w in subwidgets:
            tupwidget.addWidget(w)
        return tupwidget

    if origin is Union:
        args = [a for a in getattr(t, "__args__") if not isinstance(None, a)]
        if len(args) > 1:
            msg = f"Cannot convert Union with multiple types: {t}"
            raise TypeError(msg)
        # t is Optional[...]
        widget = type2Widget(args[0])
        if isinstance(widget, BoolCheckBox):
            widget.setTristate(True)
        return widget

    raise TypeError("Unknown type or annotation: %s" % t)


def dataclass2Widget(
    dcls: Type[DataclassProtocol],
    field_converter: Callable[[Any], FieldWidgetProtocol] = type2Widget,
    orientation: QtCore.Qt.Orientation = QtCore.Qt.Orientation.Vertical,
    globalns: Optional[Dict] = None,
    localns: Optional[Dict] = None,
    include_extras: bool = False,
) -> DataWidget:
    """
    Construct :class:`DataWidget` from *dcls*.

    Each field of *dcls* is converted to field widget by :func:`type2Widget`
    with the type hint of the field. If the field has ``Qt_typehint`` metadata,
    its value is used instead of the type hint.

    If the field type is dataclass, construction is recursively done with same
    parameters.

    Parameters
    ==========

    dcls
        Dataclass type which will be converted to widget.

    field_converter
        Callable to convert non-dataclass fields to widgets. Default is
        :class:`type2Widget`.

    orientation
        Argument for :class:`DataWidget`.

    globalns, localns, include_extras
        Arguments for :func:`get_type_hints` to resolve the forward-referenced
        type annotations.

    """
    widget = DataWidget(orientation)
    fields = dataclasses.fields(dcls)
    annots = get_type_hints(dcls, globalns, localns, include_extras)

    for f in fields:
        typehint = f.metadata.get("Qt_typehint", annots[f.name])
        if dataclasses.is_dataclass(typehint):
            field_w = dataclass2Widget(
                typehint,
                field_converter,
                orientation,
                globalns,
                localns,
                include_extras,
            )
        else:
            field_w = field_converter(typehint)  # type: ignore[assignment]
        field_w.setFieldName(f.name)
        widget.addWidget(field_w)
    return widget



from dawiq import DataclassStackedWidget, DataclassTabWidget, dataclass2Widget
from dawiq.qt_compat import QtWidgets
import dataclasses
import pytest


@dataclasses.dataclass
class DataClass1:
    x: bool


@dataclasses.dataclass
class DataClass2:
    a: bool
    b: bool
    c: bool
    d: bool
    e: bool


@dataclasses.dataclass
class DataClass3:
    x: bool


@pytest.fixture
def dataclassStackedWidget(qtbot):
    widget = DataclassStackedWidget()
    widget.addWidget(QtWidgets.QWidget())
    widget.addDataWidget(dataclass2Widget(DataClass1), DataClass1)
    widget.insertDataWidget(1, dataclass2Widget(DataClass2), DataClass2)
    return widget


def test_DataclassStackedWidget_currentDataclass(qtbot, dataclassStackedWidget):
    dataclassStackedWidget.setCurrentIndex(0)
    assert dataclassStackedWidget.currentDataclass() is None
    dataclassStackedWidget.setCurrentIndex(1)
    assert dataclassStackedWidget.currentDataclass() == DataClass2
    dataclassStackedWidget.setCurrentIndex(2)
    assert dataclassStackedWidget.currentDataclass() == DataClass1


def test_DataclassStackedWidget_indexOfDataclass(qtbot, dataclassStackedWidget):
    assert dataclassStackedWidget.indexOfDataclass(DataClass1) == 2
    assert dataclassStackedWidget.indexOfDataclass(DataClass2) == 1
    assert dataclassStackedWidget.indexOfDataclass(DataClass3) == -1


def test_DataclassStackedWidget_currentDataValueChanged(qtbot, dataclassStackedWidget):
    dataclassStackedWidget.setCurrentIndex(0)
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataValueChanged):
        dataclassStackedWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataValueChanged):
        dataclassStackedWidget.widget(2).widget(0).click()

    dataclassStackedWidget.setCurrentIndex(1)
    with qtbot.waitSignal(dataclassStackedWidget.currentDataValueChanged):
        dataclassStackedWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataValueChanged):
        dataclassStackedWidget.widget(2).widget(0).click()

    dataclassStackedWidget.setCurrentIndex(2)
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataValueChanged):
        dataclassStackedWidget.widget(1).widget(0).click()
    with qtbot.waitSignal(dataclassStackedWidget.currentDataValueChanged):
        dataclassStackedWidget.widget(2).widget(0).click()

    dataclassStackedWidget.setCurrentWidget(dataclassStackedWidget.widget(0))
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataValueChanged):
        dataclassStackedWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataValueChanged):
        dataclassStackedWidget.widget(2).widget(0).click()

    dataclassStackedWidget.setCurrentWidget(dataclassStackedWidget.widget(1))
    with qtbot.waitSignal(dataclassStackedWidget.currentDataValueChanged):
        dataclassStackedWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataValueChanged):
        dataclassStackedWidget.widget(2).widget(0).click()

    dataclassStackedWidget.setCurrentWidget(dataclassStackedWidget.widget(2))
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataValueChanged):
        dataclassStackedWidget.widget(1).widget(0).click()
    with qtbot.waitSignal(dataclassStackedWidget.currentDataValueChanged):
        dataclassStackedWidget.widget(2).widget(0).click()


def test_DataclassStackedWidget_currentDataEdited(qtbot, dataclassStackedWidget):
    dataclassStackedWidget.setCurrentIndex(0)
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataEdited):
        dataclassStackedWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataEdited):
        dataclassStackedWidget.widget(2).widget(0).click()

    dataclassStackedWidget.setCurrentIndex(1)
    with qtbot.waitSignal(dataclassStackedWidget.currentDataEdited):
        dataclassStackedWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataEdited):
        dataclassStackedWidget.widget(2).widget(0).click()

    dataclassStackedWidget.setCurrentIndex(2)
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataEdited):
        dataclassStackedWidget.widget(1).widget(0).click()
    with qtbot.waitSignal(dataclassStackedWidget.currentDataEdited):
        dataclassStackedWidget.widget(2).widget(0).click()

    dataclassStackedWidget.setCurrentWidget(dataclassStackedWidget.widget(0))
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataEdited):
        dataclassStackedWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataEdited):
        dataclassStackedWidget.widget(2).widget(0).click()

    dataclassStackedWidget.setCurrentWidget(dataclassStackedWidget.widget(1))
    with qtbot.waitSignal(dataclassStackedWidget.currentDataEdited):
        dataclassStackedWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataEdited):
        dataclassStackedWidget.widget(2).widget(0).click()

    dataclassStackedWidget.setCurrentWidget(dataclassStackedWidget.widget(2))
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataEdited):
        dataclassStackedWidget.widget(1).widget(0).click()
    with qtbot.waitSignal(dataclassStackedWidget.currentDataEdited):
        dataclassStackedWidget.widget(2).widget(0).click()


def test_DataclassStackedWidget_removeWidget(qtbot, dataclassStackedWidget):
    dataclassStackedWidget.setCurrentIndex(2)
    oldWidget = dataclassStackedWidget.currentWidget()
    dataclassStackedWidget.removeWidget(oldWidget)
    assert dataclassStackedWidget.indexOfDataclass(DataClass2) == 1
    assert dataclassStackedWidget.currentIndex() == 1
    with qtbot.waitSignals(
        [
            dataclassStackedWidget.currentDataValueChanged,
            dataclassStackedWidget.currentDataEdited,
        ]
    ):
        dataclassStackedWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataValueChanged):
        oldWidget.widget(0).click()
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataEdited):
        oldWidget.widget(0).click()

    oldWidget = dataclassStackedWidget.currentWidget()
    dataclassStackedWidget.removeWidget(oldWidget)
    assert dataclassStackedWidget.indexOfDataclass(DataClass1) == -1
    assert dataclassStackedWidget.currentIndex() == 0
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataValueChanged):
        oldWidget.widget(0).click()
    with qtbot.assertNotEmitted(dataclassStackedWidget.currentDataEdited):
        oldWidget.widget(0).click()

    oldWidget = dataclassStackedWidget.currentWidget()
    dataclassStackedWidget.removeWidget(oldWidget)
    assert dataclassStackedWidget.currentIndex() == -1


@pytest.fixture
def dataclassTabWidget(qtbot):
    widget = DataclassTabWidget()
    widget.addTab(QtWidgets.QWidget(), "EmptyTab")
    widget.addDataWidget(dataclass2Widget(DataClass1), DataClass1, DataClass1.__name__)
    widget.insertDataWidget(
        1, dataclass2Widget(DataClass2), DataClass2, DataClass2.__name__
    )
    return widget


def test_DataclassTabWidget_currentDataclass(qtbot, dataclassTabWidget):
    dataclassTabWidget.setCurrentIndex(0)
    assert dataclassTabWidget.currentDataclass() is None
    dataclassTabWidget.setCurrentIndex(1)
    assert dataclassTabWidget.currentDataclass() == DataClass2
    dataclassTabWidget.setCurrentIndex(2)
    assert dataclassTabWidget.currentDataclass() == DataClass1


def test_DataclassTabWidget_indexOfDataclass(qtbot, dataclassTabWidget):
    assert dataclassTabWidget.indexOfDataclass(DataClass1) == 2
    assert dataclassTabWidget.indexOfDataclass(DataClass2) == 1
    assert dataclassTabWidget.indexOfDataclass(DataClass3) == -1


def test_DataclassTabWidget_currentDataValueChanged(qtbot, dataclassTabWidget):
    dataclassTabWidget.setCurrentIndex(0)
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataValueChanged):
        dataclassTabWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataValueChanged):
        dataclassTabWidget.widget(2).widget(0).click()

    dataclassTabWidget.setCurrentIndex(1)
    with qtbot.waitSignal(dataclassTabWidget.currentDataValueChanged):
        dataclassTabWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataValueChanged):
        dataclassTabWidget.widget(2).widget(0).click()

    dataclassTabWidget.setCurrentIndex(2)
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataValueChanged):
        dataclassTabWidget.widget(1).widget(0).click()
    with qtbot.waitSignal(dataclassTabWidget.currentDataValueChanged):
        dataclassTabWidget.widget(2).widget(0).click()

    dataclassTabWidget.setCurrentWidget(dataclassTabWidget.widget(0))
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataValueChanged):
        dataclassTabWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataValueChanged):
        dataclassTabWidget.widget(2).widget(0).click()

    dataclassTabWidget.setCurrentWidget(dataclassTabWidget.widget(1))
    with qtbot.waitSignal(dataclassTabWidget.currentDataValueChanged):
        dataclassTabWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataValueChanged):
        dataclassTabWidget.widget(2).widget(0).click()

    dataclassTabWidget.setCurrentWidget(dataclassTabWidget.widget(2))
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataValueChanged):
        dataclassTabWidget.widget(1).widget(0).click()
    with qtbot.waitSignal(dataclassTabWidget.currentDataValueChanged):
        dataclassTabWidget.widget(2).widget(0).click()


def test_DataclassTabWidget_currentDataEdited(qtbot, dataclassTabWidget):
    dataclassTabWidget.setCurrentIndex(0)
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataEdited):
        dataclassTabWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataEdited):
        dataclassTabWidget.widget(2).widget(0).click()

    dataclassTabWidget.setCurrentIndex(1)
    with qtbot.waitSignal(dataclassTabWidget.currentDataEdited):
        dataclassTabWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataEdited):
        dataclassTabWidget.widget(2).widget(0).click()

    dataclassTabWidget.setCurrentIndex(2)
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataEdited):
        dataclassTabWidget.widget(1).widget(0).click()
    with qtbot.waitSignal(dataclassTabWidget.currentDataEdited):
        dataclassTabWidget.widget(2).widget(0).click()

    dataclassTabWidget.setCurrentWidget(dataclassTabWidget.widget(0))
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataEdited):
        dataclassTabWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataEdited):
        dataclassTabWidget.widget(2).widget(0).click()

    dataclassTabWidget.setCurrentWidget(dataclassTabWidget.widget(1))
    with qtbot.waitSignal(dataclassTabWidget.currentDataEdited):
        dataclassTabWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataEdited):
        dataclassTabWidget.widget(2).widget(0).click()

    dataclassTabWidget.setCurrentWidget(dataclassTabWidget.widget(2))
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataEdited):
        dataclassTabWidget.widget(1).widget(0).click()
    with qtbot.waitSignal(dataclassTabWidget.currentDataEdited):
        dataclassTabWidget.widget(2).widget(0).click()


def test_DataclassTabWidget_removeTab(qtbot, dataclassTabWidget):
    dataclassTabWidget.setCurrentIndex(2)
    oldTab = dataclassTabWidget.currentWidget()
    dataclassTabWidget.removeTab(dataclassTabWidget.indexOf(oldTab))
    assert dataclassTabWidget.currentIndex() == 1
    with qtbot.waitSignals(
        [
            dataclassTabWidget.currentDataValueChanged,
            dataclassTabWidget.currentDataEdited,
        ]
    ):
        dataclassTabWidget.widget(1).widget(0).click()
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataValueChanged):
        oldTab.widget(0).click()
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataEdited):
        oldTab.widget(0).click()

    oldTab = dataclassTabWidget.currentWidget()
    dataclassTabWidget.removeTab(dataclassTabWidget.indexOf(oldTab))
    assert dataclassTabWidget.currentIndex() == 0
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataValueChanged):
        oldTab.widget(0).click()
    with qtbot.assertNotEmitted(dataclassTabWidget.currentDataEdited):
        oldTab.widget(0).click()

    oldTab = dataclassTabWidget.currentWidget()
    dataclassTabWidget.removeTab(dataclassTabWidget.indexOf(oldTab))
    assert dataclassTabWidget.currentIndex() == -1
